import torch,os
from torch import nn
import numpy as np
from .Records import BatchRecord
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')

"""
Used to apply training updates to the Multi Input Model. May also output graphs
of the training loss
"""

class MultiInputTrainer():
	def __init__(self,
		model,
		lr=1e-5,
		betas = (0.5,0.999),
		loss_function = nn.MSELoss(),
		batch_size = 64,
		regress = True,
		loss_image_dir = None,
		checkpoint_dir = None,
		name = 'experiment_name',
		verbose = False,
		save_latest_freq = 100):
		
		# Core variables for training the self.model
		self.model = model
		self.name = name

		self.optimizer = torch.optim.Adam(
			self.model.classifier_parameters(),
			betas = betas,
			lr= lr
		)
		self.optimizer_reg = torch.optim.Adam(
			self.model.regressor.parameters(),
			betas = betas,
			lr = lr
		)

		# Load saved or pretrained models
		self.checkpoint_dir = checkpoint_dir
		if self.checkpoint_dir is not None:
			os.makedirs(self.checkpoint_dir,exist_ok=True)
			self.model_file = os.path.join(
				self.checkpoint_dir,
				'%s.pt' % self.name)
			if os.path.isfile(self.model_file):
				state_dicts = torch.load(self.model_file)
				self.model.load_state_dict(state_dicts['model_state_dict'])
				self.model.regressor_freeze()
				self.optimizer.load_state_dict(
					state_dicts['optimizer_state_dict']
				)
				self.optimizer_reg.load_state_dict(
					state_dicts['optimizer_reg_state_dict']
				)
		self.loss_function = loss_function
		self.batch_size = batch_size
		self.regress = regress
		self.index = 0
		self.one_step = True
				
		# Outputs images of the loss function over time
		self.loss_image_dir = loss_image_dir
		if self.loss_image_dir is not None:
			os.makedirs(self.loss_image_dir,exist_ok=True)
			self.loss_image_file = os.path.join(self.loss_image_dir,
											f"{self.name}_loss.png")
			self.loss_vals_file  = os.path.join(self.loss_image_dir,
											f"{self.name}_vals.npy")
			self.pids_read = set()
			self.x_files_read = set()
			if os.path.isfile(self.loss_vals_file):
				_ = np.load(self.loss_vals_file)
				self.xs,self.ys,self.ys_2,self.ys_c_dud = \
						[list(_[i,:]) for i in range(4)]
				if self.model.variational:
					self.ys_kl = list(_[i,4])
	
				self.loss_Y          = self.ys[-1]
				self.loss_C_dud      = self.ys_c_dud[-1]
				self.loss_regressor  = self.ys_2[-1]
				if model.variational:
					self.loss_kl = self.ys_kl[-1]
				self.loss_classifier = self.loss_Y + self.loss_C_dud
			else:
				self.xs,self.ys,self.ys_2,self.ys_kl,self.ys_c_dud =\
							[[] for _ in range(5)]
		
				self.loss_Y          = 0
				self.loss_C_dud      = 0
				self.loss_classifier = 0
				self.loss_regressor  = 0
				if model.variational:
					self.loss_kl = 0
			
		self.verbose = verbose
		self.save_latest_freq = save_latest_freq
		
	def loop(self, pr: BatchRecord, dataloader = None):
		assert(isinstance(pr,BatchRecord))
		x = self.model(pr,return_encoded=True)
		y_pred,y_reg = self.model(x,
				encoded_input=True,
				dates=pr.get_exam_dates(),
				bdate=pr.get_birth_dates()[0])
		if self.one_step:
			self.loss_Y = self.loss_function(y_pred,pr.get_Y())
			self.loss_C_dud = self.loss_function(y_reg,pr.get_C_dud())
			self.loss_classifier = self.loss_Y + (self.loss_C_dud)
			if self.model.variational:
				self.loss_kl = self.model.kl * 0.0005
				self.loss_classifier = self.loss_classifier + loss_kl
			self.loss_classifier.backward()
		else:
			if self.regress:
				self.loss_regressor = \
					self.loss_function(y_reg,pr.get_C())
			else:
				self.loss_regressor = \
					self.loss_function(y_reg,pr.get_C_dud())
			self.loss_regressor.backward()
		
		if self.loss_image_dir is not None:
			
			self.ys.append(float(self.loss_Y))
			self.ys_2.append(float(self.loss_regressor))
			self.xs.append(self.index)
			if self.model.variational:
				self.ys_kl.append(float(self.loss_kl))
			self.ys_c_dud.append(float(self.loss_C_dud))

		if self.verbose:
			if self.model.variational:
				print(("%d: Class: %.6f, KL: %.6f, Dud: "+\
					"%.6f, Reg: %.6f (%d, %d) | %s") % \
					(i,float(loss_Y), float(loss_kl), float(loss_C_dud),
					float(loss_regressor),len(self.pids_read),
					len(self.x_files_read),self.name))
			else:
				print("%d: Class: %.6f, Dud: %.6f, Reg: %.6f (%d, %d) | %s" % \
					(i,float(loss_Y), float(loss_C_dud),
					float(loss_regressor),len(self.pids_read),
					len(self.x_files_read),self.name))

		if self.index % self.batch_size == 0 and self.index != 0:
			if self.one_step:
				self.optimizer.step()
				self.optimizer.zero_grad()
				self.model.classifier_freeze()
				if dataloader is not None:
					dataloader.switch_stack()
			else:
				self.optimizer_reg.step()
				self.optimizer_reg.zero_grad()
				self.model.regressor_freeze()
				if dataloader is not None:
					dataloader.switch_stack()
			self.one_step = not self.one_step
		if self.index % self.save_latest_freq == 0 and self.index != 0:
			if self.checkpoint_dir is not None:
				torch.save({
					'model_state_dict' : self.model.state_dict(),
					'optimizer_state_dict' : self.optimizer.state_dict(),
					'optimizer_reg_state_dict' : self.optimizer_reg.state_dict()
					},
					self.model_file
				)
			if self.loss_image_dir is not None:
				plt.plot(list(range(len(self.xs))),self.ys,
					label="Classifier loss - label")
				if self.regress:
					plt.plot(list(range(len(self.xs))),
						self.ys_2,
						label="Regressor loss")
				if self.model.variational:
					plt.plot(list(range(len(self.xs))),
						self.ys_kl,
						label="KL Loss")
				plt.plot(list(range(len(self.xs))),
					self.ys_c_dud,label="Classifier loss - adversarial")
				plt.legend()
				plt.savefig(self.loss_image_file)
				plt.clf()
				
				if self.model.variational:
					np.save(self.loss_vals_file,np.array([self.xs,
													self.ys,
													self.ys_2,
													self.ys_c_dud,
													self.ys_kl]))
				else:
					np.save(self.loss_vals_file,np.array([self.xs,
													self.ys,
													self.ys_2,
													self.ys_c_dud]))
				
		self.index += 1

	def test(self):
		return



#!/usr/bin/python3
import os,sys
import torchvision.transforms as transforms
from pdb import set_trace as st
import random
import numpy as np
import torch
import pandas as pd
import torch.multiprocessing
from .utils import *
torch.multiprocessing.set_sharing_strategy('file_system')
import psutil,shutil
import nibabel as nb
from nibabel.filebasedimages import *
import gc
from .Records import BatchRecord,ImageRecord
from .DataBaseWrapper import DataBaseWrapper

# Translates a filename to a key and back, for storing files as keys in the
# pandas dataframe. By default, the keys are the full filepaths. This function
# may need to be changed when switching to different systems
def key_to_filename_default(fkey,reverse=False):
	return fkey

class MedImageLoader():
	def __init__(self,*image_folders,
			pandas_cache = '../pandas/',
			cache = True,
			key_to_filename = key_to_filename_default,
			batch_by_pid=True,
			file_record_name=None,
			all_vars = None,
			batch_size = 14,
			dim = (96,96,96),
			get_encoded = False,
			static_inputs = None,
			confounds = [],
			match_confounds = [],
			label = [],
			group_by = None,
			augment = True,
			val_ranges = {},
			dtype="torch",
			Y_dim = (1,32),
			C_dim = (16,32),
			return_obj = False,
			channels_first = True,
			recycle=True,
			gpu_ids = "",
			save_ram = True):
		self.channels_first = channels_first
		self.image_folders = image_folders
		self.augment = augment
		self.dim=dim
		self.dtype=dtype
		self.cache = cache
		self.pandas_cache = pandas_cache
		self.val_ranges = val_ranges
		self.get_encoded = get_encoded
		self.batch_by_pid = group_by is not None
		self.file_list_dict = {}
		self.static_inputs = static_inputs
		self.label = label
		self.confounds = confounds
		self.group_by = group_by
		self.Y_dim = Y_dim
		self.C_dim = C_dim
		self.mode = None
		self.gpu_ids = gpu_ids
		
		# If set to true, restacks images every time via the data 
		# matching function. Best for very large and imbalanced datasets
		self.recycle=recycle
		
		check_key_to_filename(key_to_filename)

		# Stores images so that they aren't repeated in different stacks
		self.image_dict = {}
		# If true, this uses one match confound at a time and cycles through
		# them
		self.zero_data_list = []
		self.match_confounds = match_confounds
		if self.batch_by_pid:
			self.batch_size = 1
		else:
			self.batch_size = batch_size
		self.return_obj = return_obj
		
		# Create or read in the image database
		if self.pickle_input():
			self.all_vars_file = self.image_folders[0]
			#assert(self.cache or os.path.isfile(self.all_vars_file))
		else:
			os.makedirs(os.path.dirname(self.pandas_cache),
				exist_ok=True)
			self.all_vars_file = os.path.join(
				self.pandas_cache,
				"all_vars_%s.pkl" % get_dim_str(dim=self.dim))
		
		self.all_vars = DataBaseWrapper(
					filename=self.all_vars_file,
					labels=self.label,
					confounds=self.confounds,
					dim=self.dim,
					key_to_filename=key_to_filename,
					val_ranges=self.val_ranges)
		if not self.pickle_input():
			self.build_pandas_database()
		self.all_vars.build_metadata()
		
		# Determine which mode the scheduler is in
		if (self.label is not None and len(self.label) > 0):
			self.mode = "match"
		else:
			self.mode = "iterate"
		
		self.mem_limit = psutil.virtual_memory().available * 0.2
		
		# For outputting the records of files that were read in
		self.file_record_name = file_record_name 
		if self.file_record_name is not None:
			self.file_record = {}
			fd = os.path.join(wd,'json','dataloader_records')
			if not os.path.isdir(fd): os.makedirs(fd)
			self.file_record_output = os.path.join(fd,file_record_name)

		self.uniques = None
		self.n_buckets = 3
		self.rmatch_confounds = self.confounds
	
		# Regular stack variables
		self.file_list_dict = {}
	
		# Switch stack variables
		self.file_list_dict_hidden = {}
		self.match_confounds_hidden = []
		self.mode_hidden = "iterate"
		if self.return_labels():
			for l in self.label:
				if l not in self.val_ranges:
					self.match_confounds_hidden.append(l)
		self.index = 0
		self.save_ram = save_ram
		self.load_image_stack()
	# Builds up the entire cache in one go — may take a while
	def build_pandas_database(self):
		assert(self.all_vars is not None)
		assert(not self.pickle_input())
		old_mode = self.mode
		self.mode = "iterate"
		self._load_list_stack()
		
		for i,filename in enumerate(self.image_dict):
			im = self.image_dict[filename]
			if not self.all_vars.has_im(im):
				try:
					im.get_image()
					im.clear_image()
				except ImageFileError:
					continue
			if i % 100 == 0:
				self.all_vars.out_dataframe()
		self.clear_mem()
		self.all_vars.out_dataframe()
		self.mode = old_mode
	def pickle_input(self):
		return len(self.image_folders) == 1 and \
			os.path.splitext(self.image_folders[0])[1] == ".pkl"
	def tl(self):
		# Top label
		
		if len(self.label) == 0: tl = "Folder"
		else: tl = self.label[0]
		if tl not in self.file_list_dict:
			self.file_list_dict[tl] = []
		return tl
	def get_file_list(self):
		if self.pickle_input() and self.tl() == "Folder":
			return [[str(_) for _ in self.all_vars.get_file_list()]]
		if self.mode == "iterate":
			if self.pickle_input():
				fname_list = [str(_) for _ in self.all_vars.get_file_list()]
				return self.all_vars.stack_list_by_label(fname_list,self.tl())
			else:
				all_filename_lists = []
				duplicate_test = set()
				for img in self.image_folders:
					flist = get_file_list(img)#,db_builder=self.all_vars)
					flist_set = set(flist)
					if len(flist_set.intersection(duplicate_test))>0:
						raise Exception(
							"Intersecting files found between labels"
							)
					duplicate_test = duplicate_test.union(flist_set)
					all_filename_lists.append(flist)
				assert(len(all_filename_lists) > 0)
				return all_filename_lists
		elif self.mode == "match":
			[fname_list],_ = get_balanced_filename_list(self.tl(),
				self.match_confounds,
				selection_ratios=[1], total_size_limit=np.inf,
				non_confound_value_ranges = self.val_ranges,verbose=False,
				all_vars=self.all_vars.all_vars)
			fname_list = list(fname_list)
			fname_list = [self.all_vars.key_to_filename(_) for _ in fname_list]
			if len(fname_list) == 0:
				print(self.all_vars.all_vars.loc[:,self.tl()])
				raise Exception("No valid files from %s" % self.tl())
			assert(isinstance(fname_list,list))
			return self.all_vars.stack_list_by_label(fname_list,self.tl())
		else:
			raise Exception("Invalid mode: %s" % self.mode)
	def _load_list_stack(self):
		X_files = self.get_file_list() 
		for i,filename_list in enumerate(X_files):
			for j,filename in enumerate(filename_list):
				if filename in self.image_dict:
					X_files[i][j] = self.image_dict[filename]
				else:
					X_files[i][j] = ImageRecord(filename,
								dim=self.dim,
								y_nums=[i] if len(X_files) == 1 else None,
								Y_dim = self.Y_dim,
								C_dim = self.C_dim,
								dtype=self.dtype,
								all_vars = self.all_vars,
								cache=self.cache,
								static_inputs = self.static_inputs)
					self.image_dict[filename] = X_files[i][j]
		return X_files
	def load_image_stack(self):
		if self.get_mem() > self.mem_limit:
			self.clear_mem()
		#self.rotate_labels()
		if self.tl() not in self.file_list_dict:
			self.file_list_dict[self.tl()] = []
		X_files = self._load_list_stack()
		if self.batch_by_pid:
			pdict = {}
			for images in X_files:
				for image in images:
					image.load_extra_info()
					if not is_nan(image.group_by):
						if image.group_by not in pdict:
							pdict[image.group_by] = []
						pdict[image.group_by].append(image)
			image_record_list = []
			for ID in pdict:
				image_record_list.append(
							BatchRecord(pdict[ID],
							batch_by_pid=True,
							sort=True,
							dtype=self.dtype,
							channels_first=self.channels_first))
			X_files = [image_record_list]
		self.file_list_dict[self.tl()] = X_files
	def return_labels(self):
		if self.tl() == "Folder":
			if len(self.image_folders) > 1:
				return True
			elif self.pickle_input():
				return False
		return self.label is not None and len(self.label) > 0
	def record(self,flist,index=None):
		self.file_record = set(self.file_record).union(set(flist))
		if index is None or index % 100 == 0:
			with open(self.file_record_output,"w") as fileobj:
				json.dump(list(self.file_record),fileobj,indent=4)
	def read_record(self):
		if os.path.isfile(self.file_record_output):
			with open(self.file_record_output,"r") as fileobj:
				self.file_record = set(json.load(fileobj))
	def rotate_labels(self,zero_list_addendum = None):
		if len(self.label) < 2: return
		if zero_list_addendum is not None:
			self.zero_data_list.append(zero_list_addendum)
		self.label = self.label[1:] + [self.label[0]]
		while self.label[0] in self.zero_data_list:
			self.label = self.label[1:] + [self.label[0]]
	def switch_stack(self):
		self.label,self.rmatch_confounds = self.rmatch_confounds,self.label
		self.file_list_dict,self.file_list_dict_hidden = \
			self.file_list_dict_hidden,self.file_list_dict
		self.match_confounds,self.match_confounds_hidden = \
			self.match_confounds_hidden,self.match_confounds
		self.mode,self.mode_hidden = self.mode_hidden,self.mode
	def __next__(self):
		if len(self) == 0: self.load_image_stack()
		if self.index > len(self):
			self.all_vars.out_dataframe()
			self.index = 0
			self.rotate_labels()
			self.load_image_stack()
			raise StopIteration
		# Temporary measure
		if self.index % 1000 == 0 and self.index != 0:
			self.clear_mem()
		temp = []
		for i in range(self.batch_size):
			b = i % len(self.file_list_dict[self.tl()])
			img = None
			if len(self.file_list_dict[self.tl()][b]) == 0: continue
			for j in range(len(self.file_list_dict[self.tl()][b])):
				im = self.file_list_dict[self.tl()][b].pop()
				self.file_list_dict[self.tl()][b] = [im] + \
					self.file_list_dict[self.tl()][b]
				try:
					if self.cache: assert self.all_vars is not None
					img = im.get_image(augment=self.augment)
					temp.append(im)
					self.index += 1
					break
				except (ImageFileError, ValueError) as e:
					self.file_list_dict[self.tl()][b] = \
						self.file_list_dict[self.tl()][b][1:]
					continue
		if len(temp) != self.batch_size:
			print(len(temp))
			print(temp)
			print(self.batch_size)
			print(self.file_list_dict)
		assert(len(temp) == self.batch_size)
		if self.batch_by_pid:
			p = temp[0]
		else:
			p = BatchRecord(temp,
				dtype = self.dtype,
				sort=False,
				batch_by_pid=False,
				channels_first=self.channels_first,
				gpu_ids=self.gpu_ids)
		if self.return_obj:
			return p
		elif self.return_labels():
			return p.get_image(),p.get_Y()
		else:
			return p.get_image()				

	def __len__(self):
		l = len(self.file_list_dict[self.tl()])
		if l == 0:
			return 0
		else:
			return l * max([len(_) for _ in self.file_list_dict[self.tl()]])
	def __iter__(self):
		return self
	def clear_mem(self):
		cur_mem = self.get_mem()
		times_called = []
		ssize = 0
		for filename in self.image_dict:
			if self.image_dict[filename].image is not None:
				times_called.append(self.image_dict[filename].times_called)
				#if ssize == 0: ssize = self.image_dict[filename].get_mem()
		times_called = sorted(times_called,reverse=True)
		#n = int(self.mem_limit / ssize * 0.5)
		#if n > len(times_called):
		#	warnings.warn("Something strange")
		#	n = int(len(times_called)/2)
		n = 10
		median_times_called = 1000000000 if len(times_called) < n else times_called[n]
		for filename in self.image_dict:
			if self.image_dict[filename].image is not None:
				if self.save_ram or self.image_dict[filename].times_called < median_times_called:
					self.image_dict[filename].clear_image()
		if self.save_ram:
			del self.image_dict
			self.image_dict = {}
			gc.collect()
	def get_mem(self):
		total_mem = 0
		for filename in self.image_dict:
			total_mem += float(self.image_dict[filename].get_mem())
		return total_mem
	def name(self):
		return 'MedImageLoader'



import functools,os
import numpy as np
from monai.transforms import *
from .utils import *

generate_transforms = Compose([
		RandAffine(prob=0.5, translate_range=10), 
		RandRotate(prob=0.5, range_x=10.0),
		RandGaussianNoise(prob=0.5),
		RandBiasField(prob=0.5)])
#,
#		RandSmoothDeform(spatial_size=(96,96,96),rand_size=(96,96,96))
#])

@functools.total_ordering
class ImageRecord():
	def __init__(self,
		filename,
		all_vars = None,
		dim=(96,96,96),
		dtype="torch",
		ID = None,
		extra_info_list = None,
		y_on_c = True,
		cache = True,
		Y_dim = (1,32),
		C_dim = (16,32),
		y_nums = None,
		c_nums = None,
		static_inputs=[]):
		
		self.dim=dim
		self.filename = filename
		self.npy_file = get_dim_str(self.filename,self.dim)
		self.image_type = None
		self.dtype = dtype
		self.all_vars = all_vars
		self.Y_dim = Y_dim
		self.C_dim = C_dim
		self.y_on_c = y_on_c # Adds the Y value to the C array as well
		
		self.is_loaded = False			
		self.group_by = None
		self.bdate = None
		self.exam_date = datetime.datetime(year=1970,month=1,day=1)
		self.ID = None
		self.extra_info_list = None
		self.json_file = None
		
		self.image = None
		self.Y = None
		self.C = None
		self.y_nums = y_nums
		self.c_nums = c_nums
		
		self.static_inputs = []
		self.static_input_res = None
		
		self.loaded = False
		if self.all_vars is not None:
			if self.npy_file in self.all_vars.all_vars:
				self.load_extra_info()
		
		self.times_called = 0
		self.cache = cache
		self.cached_record = None
	def get_exam_date(self):
		if not self.loaded:
			self.load_extra_info()
		return self.exam_date
	def get_birth_date(self):
		if not self.loaded:
			self.load_extra_info()
		return self.bdate
	def load_extra_info(self):
		if self.loaded: return
		if self.all_vars is None:
			return
		if not self.all_vars.has_im(self):
			return
		self.bdate = self.all_vars.get_birth_date(self.npy_file)
		self.exam_date = self.all_vars.get_exam_date(self.npy_file)
		self.group_by = self.all_vars.get_ID(self.npy_file)
		self.loaded = True
	def get_static_inputs(self):
		if self.static_input_res is None:
			self.static_input_res = []
			for key in self.static_inputs:
				static_inputs.append(
					self.all_vars.loc_val(self.npy_file,key)
				)
		return self.static_input_res
	def _is_valid_operand(self, other):
		return hasattr(other, "exam_date") and hasattr(other,"group_by")
	def __eq__(self, other):
		if not self._is_valid_operand(other):
			return NotImplemented
		return (self.exam_date == other.exam_date)
	def __lt__(self, other):
		if not self._is_valid_operand(other):
			return NotImplemented
		return self.exam_date < other.exam_date
	def get_image_type(self):
		if self.image_type is None:
			if os.path.isdir(self.filename):
				self.image_type = "dicom_folder"
			elif os.path.isfile(self.filename):
				name,ext = os.path.splitext(self.filename)
				name = name.lower()
				ext = ext.lower()
				if ext == ".npy":
					self.image_type = "npy"
				elif ext  == ".nii":
					self.image_type = "nifti"
				elif ext == ".gz" and os.path.splitext(name)[1] == ".nii":
					self.image_type = "nifti"
				elif ext == ".dcm":
					self.image_type = "dicom"
				else:
					raise Exception(
					"Not implemented for extension %s" % ext)
			elif os.path.isfile(
				get_dim_str(self.filename,
						dim=self.dim,
						outtype='.nii.gz')):
				self.filename = get_dim_str(self.filename,
							dim=self.dim,
							outtype='.nii.gz')
				return self.get_image_type()
			elif os.path.isfile(
				get_dim_str(self.filename,
						dim=self.dim,
						outtype='.nii')):
				self.filename = get_dim_str(self.filename,
							dim=self.dim,
							outtype='.nii')
				return self.get_image_type()
			elif os.path.isdir(get_dim_str(self.filename,
							dim=self.dim,
							outtype='dicom')):
				self.filename = get_dim_str(self.filename,
							dim=self.dim,
							outtype='dicom')
				return self.get_image_type()
		return self.image_type
	def get_mem(self):
		if self.image is None:
			return 0
		#else:
		#	return np.prod(self.image.shape) * \
		#		np.dtype(self.image.dtype).itemsize
		elif self.dtype == "torch":
			return self.image.element_size() * self.image.nelement()
		elif self.dtype == "numpy":
			return np.prod(self.image.shape) * \
				np.dtype(self.image.dtype).itemsize
		#else:
		#	raise Exception("Unimplemented dtype: %s" % self.dtype)
	def clear_image(self):
		self.image = None
		self.X = None
		self.Y = None
		self.C = None
		#self.loaded = False
	def read_image(self):
		if self.get_image_type() == "dicom_folder":
			self.filename,self.json_file = compile_dicom(self.filename,
				db_builder=self.all_vars)
			self.image_type = None
		self.cached_record = get_dim_str(self.filename,self.dim)
		if self.cache and os.path.isfile(self.cached_record):
			self.image = np.load(self.cached_record)
		elif self.get_image_type() == "nifti":
			self.image = nb.load(self.filename).get_fdata()
		elif self.get_image_type() == "npy":
			self.image = np.load(self.filename)
		elif self.get_image_type() == "dicom":
			self.image = dicom.dcmread(self.filename).pixel_array
		else:
			print("Error in %s" % self.filename)
			print("Error in %s" % self.npy_file)
			raise Exception("Unsupported image type: %s"%self.get_image_type())
		self.image = resize_np(self.image,self.dim)
		if self.cache and not os.path.isfile(self.cached_record):
			np.save(self.cached_record,self.image)
		if self.all_vars is not None:
			self.all_vars.add_json(nifti_file=self.filename)
		if self.dtype == "torch":
			self.image = torch.tensor(self.image)
	def get_image(self,augment=False):
		if self.image is None:
			self.read_image()
		self.load_extra_info()
		self.times_called += 1
		if augment and self.dtype == "torch":
			return generate_transforms(self.image)
		else: return self.image
	def _get_Y(self):
		"""Returns label"""
		if self.y_nums is not None:
			return self.y_nums
		self.y_nums = self.all_vars.get_label_encode(self.npy_file)
		return self.y_nums
	def _get_C(self):
		"""Returns confound array"""
		if self.c_nums is not None:
			return self.c_nums
		self.c_nums = self.all_vars.get_confound_encode(self.npy_file)
		return self.c_nums
	def get_Y(self):
		if self.Y is not None:
			return self.Y
		y_nums = self._get_Y()
		if self.dtype == "numpy":
			self.Y = np.zeros(self.Y_dim)
		elif self.dtype == "torch":
			self.Y = torch.zeros(self.Y_dim)
		for i,j in enumerate(y_nums):
			self.Y[i,j] = 1
		return self.Y
	def get_C(self):
		if self.C is not None:
			return self.C
		c_nums = self._get_C()
		if self.dtype == "numpy":
			self.C = np.zeros(self.C_dim)
		elif self.dtype == "torch":
			self.C = torch.zeros(self.C_dim)
		for i,j in enumerate(c_nums):
			self.C[i,j] = 1
		if self.y_on_c:
			y_nums = self._get_Y()
			for i,j in enumerate(y_nums):
				self.C[i+len(self.all_vars.confounds),j] = 1
		return self.C
	def get_C_dud(self):
		if self.dtype == "numpy":
			C_dud = np.zeros(self.C_dim)
			C_dud[0,:len(self.all_vars.confounds)] = 1
		elif self.dtype == "torch":
			C_dud = torch.zeros(self.C_dim)
			C_dud[0,:len(self.all_vars.confounds)] = 1
		if self.y_on_c:
			y_nums = self._get_Y()
			for i,j in enumerate(y_nums):
				C_dud[i+len(self.all_vars.confounds),j] = 1
		return C_dud
		
class BatchRecord():
	def __init__(self,
			image_records,
			dtype="torch",
			device=None,
			sort=True,
			batch_by_pid=False,
			channels_first = True,
			gpu_ids=""):
		self.image_records = image_records
		assert(
			np.all(
				[isinstance(image_record,ImageRecord) \
					for image_record in image_records
				]
				)
			)
		self.gpu_ids = gpu_ids
		self.channels_first = channels_first
		self.batch_by_pid=batch_by_pid
		if sort: self.image_records = sorted(image_records)
		#assert(
		#	np.all(
		#		[_.group_by == self.image_records[0].group_by \
		#			for _ in self.image_records]
		#	)
		#)
		self.device=device
		self.dtype=dtype
	#def process_to_cuda(self,device):
	#	self.Y = torch.tensor(self.Y).float().cuda(device)
	#	self.Y = torch.unsqueeze(self.Y,1)
	#	self.Y = self.Y[-1,...]
	#	
	#	self.X = torch.tensor(self.X)
	#	self.X = torch.unsqueeze(self.X,1).float().cuda(device)
	#	
	#	self.C = torch.tensor(self.C).float().cuda(device)
	#	self.C_dud = torch.tensor(self.C_dud).float().cuda(device)
	
	def name(self):
		return 'BatchRecord'
	def _get(self,callback,augment=False):
		Xs = []
		no_arr = False
		for im in (self.image_records if (callback != "Y" and not self.batch_by_pid)\
			else [self.image_records[-1]]):
			if callback == "X":
				X = im.get_image(augment=augment)
				if self.dtype == "torch":
					assert(len(X.size()) == 3)
					if self.channels_first:
						X = torch.unsqueeze(X,0)
					else:
						X = torch.unsqueeze(X,-1)
				elif self.dtype == "numpy":
					assert(len(X.shape) == 3)
					if self.channels_first:
						X = np.expand_dims(X,axis=0)
					else:
						X = np.expand_dims(X,axis=-1)
				else:
					raise Exception("Invalid dtype: %s" % self.dtype)
			elif callback == "Y":
				X = im.get_Y()
			elif callback == "C":
				X = im.get_C()
			elif callback == "C_dud":
				X = im.get_C_dud()
			elif callback == "birth_dates":
				X = im.get_birth_date()
				no_arr = True
			elif callback == "exam_dates":
				X = im.get_exam_date()
				no_arr = True
			elif callback == "static_inputs":
				X = im.get_static_inputs()
				no_arr = True
			else:
				raise Exception("Invalid callback")
			if no_arr:
				Xs.append(X)
			elif self.dtype == "torch":
				Xs.append(torch.unsqueeze(X,0))
			elif self.dtype == "numpy":
				Xs.append(np.expand_dims(X,axis=0))
			else:
				raise Exception("Invalid dtype")
		if no_arr:
			return Xs
		elif self.dtype == "torch":
			Xs = torch.concatenate(Xs,0)
			if self.gpu_ids == "":
				return Xs.float()
			else:
				return Xs.float().cuda(self.gpu_ids[0])
		elif self.dtype == "numpy":
			Xs = np.concatenate(Xs,axis=0)
			return Xs #.astype(np.float32)
		else:
			raise Exception("Invalid dtype: %s" % self.dtype)
	def get_image(self,augment=False):
		return self._get("X",augment=augment)
	def get_Y(self):
		return self._get("Y")
	def get_C(self):
		return self._get("C")
	def get_C_dud(self):
		return self._get("C_dud")
	def get_exam_dates(self):
		return self._get("exam_dates")
	def get_birth_dates(self):
		return self._get("birth_dates")
	def get_static_inputs(self):
		return self._get("static_inputs")




import torch.nn as nn
from torch.autograd import Variable
import torch
import math
import random
import itertools
import numpy as np
from datetime import datetime
import hashlib
import os,sys
from copy import deepcopy as copy
from .Records import ImageRecord,BatchRecord
from .utils import download_weights

# Three functions that are used to get the age encoding functions
def time_index(i,pos,d=512,c=10000):
	if i % 2 == 0:
		v = math.sin(pos/(c**(2*i/d)))
	else:
		v = math.cos(pos/(c**(2*i/d)))
	return v

def get_age_arr(age,max_age=120.0,d=512):
	arr = np.zeros((d,))
	pos = int(age / max_age * 1000)
	for i in range(d):
		arr[i] = time_index(i,pos,d)
	return arr

def get_age_encoding(date,birthdate,d=512):
	if date is None or birthdate is None:
		return np.zeros((d,0))
	age = date.year - birthdate.year
	return get_age_arr(age,d=d)

# https://stackoverflow.com/questions/26685067/represent-a-hash-string-to-binary-in-python
# Two functions to encode strings as binary arrays
def text_to_bin(text, n_bin=32,d=512):
	if text is None: text=""
	text=text.lower()
	word_ord = '{}'.format(
			bin(int(hashlib.md5(text.encode('utf-8')).hexdigest(), n_bin))
		)
	word_ord = word_ord[2:]
	arr = []
	for i in range(d):
		a = word_ord[i % len(word_ord)]
		if a == "1":
			arr.append(1.0)
		elif a == "0":
			arr.append(0.0)
		else:
			raise Exception("%s is bad"% str(a))
	return arr

def encode_static_inputs(static_input,d=512):
	arr = np.zeros((len(static_input),d))
	for i in range(len(static_input)):
		arr[i,:] = text_to_bin(static_input[i],d=d)
	return arr

class Reshape(nn.Module):
	'''
		Used in a nn.Sequential pipeline to reshape on the fly.
	'''
	def __init__(self, *target_shape):
		super().__init__()
		self.target_shape = target_shape
	
	def forward(self, x):
		return x.view(*self.target_shape)

class Encoder(nn.Module):
	def __init__(self,latent_dim=512):
		super(Encoder,self).__init__()
		nchan=1
		base_feat = 64
		self.encoder = nn.Sequential(
			nn.Conv3d(in_channels = nchan, out_channels = base_feat, stride=2,
				kernel_size=5, padding = 2), #1*96*96*96 -> 64*48*48*48
			nn.LeakyReLU(),
			#nn.InstanceNorm3d(base_feat),
			nn.Conv3d(in_channels = base_feat, out_channels = base_feat*2,
				stride=2, kernel_size = 5,
				padding=2), #64*48*48*48 -> 128*24*24*24
			nn.LeakyReLU(),
			nn.InstanceNorm3d(base_feat*2),
			nn.Conv3d(in_channels = base_feat*2, out_channels = base_feat*4,
				stride=2,kernel_size = 3,
				padding=1), #128*24*24*24 -> 256*12*12*12
			nn.LeakyReLU(),
			#nn.InstanceNorm3d(base_feat*4),
			nn.Conv3d(in_channels = base_feat*4, out_channels = base_feat*4,
				stride=4,kernel_size = 5,padding=2), #256*12*12*12 -> 256*3*3*3
			nn.LeakyReLU(),
			nn.InstanceNorm3d(base_feat*4),
			nn.Conv3d(in_channels = base_feat*4, out_channels = base_feat*32,
				stride=1,kernel_size = 3,padding=0), #256*3*3*3 -> 2048*1*1*1
			nn.LeakyReLU(),
			Reshape([-1,base_feat*32]),
			nn.Linear(in_features = base_feat*32, out_features = base_feat*16),
			nn.LeakyReLU(),
			nn.Linear(in_features = base_feat*16, out_features = latent_dim),
			
		)
	def parameters(self):
		return self.encoder.parameters()
	def forward(self, x):
		x = self.encoder(x)
		return x


class Decoder(nn.Module):
	def __init__(self):
		super(Decoder,self).__init__()
		nchan=1
		base_feat = 64
		latent_dim = 512	
		self.decoder = nn.Sequential(
			nn.Linear(in_features = latent_dim, out_features = base_feat*16),
			nn.ReLU(),
			nn.BatchNorm1d(base_feat*16),
			nn.Linear(in_features = base_feat*16, out_features = base_feat*32),
			nn.ReLU(),
			Reshape([-1,base_feat*32,1,1,1]),
			nn.ConvTranspose3d(in_channels = base_feat*32,
				out_channels = base_feat*16, kernel_size = 3,stride=1,
				padding=0), #2048*1*1*1 -> 1024*3*3*3
			nn.ReLU(),
			nn.BatchNorm3d(base_feat*16),
			nn.ConvTranspose3d(in_channels = base_feat*16,
				out_channels = base_feat*4, kernel_size = 4,stride=2, padding=1,
				bias=False), #256*3*3*3 -> 256*6*6*6
			nn.ReLU(),
			nn.BatchNorm3d(base_feat*4),
			nn.ConvTranspose3d(in_channels = base_feat*4,
				out_channels = base_feat*4, kernel_size = 4,stride=2, padding=1,
				bias=False), #256*6*6*6 -> 256*12*12*12
			nn.ReLU(),
			nn.BatchNorm3d(base_feat*4),
			nn.ConvTranspose3d(in_channels = base_feat*4,
				out_channels = base_feat*2, kernel_size = 4,stride=2, padding=1,
				bias=False), #256*12*12*12 -> 128*24*24*24
			nn.ReLU(),
			nn.BatchNorm3d(base_feat*2),
			nn.ConvTranspose3d(in_channels = base_feat*2,
				out_channels = base_feat, kernel_size = 4, stride=2, padding=1,
				bias=False), #128*24*24*24 -> 64*48*48*48
			nn.ReLU(),
			nn.BatchNorm3d(base_feat),
			nn.ConvTranspose3d(in_channels = base_feat, out_channels = 1,
				kernel_size = 4,stride=2,padding=1,
				bias=False), #64*48*48*48 -> 1*96*96*96
			nn.Sigmoid()
		)
	def forward(self, x):
		x = self.decoder(x)
		return x

class Regressor(nn.Module):
	def __init__(self,latent_dim,n_confounds,n_choices,device='cpu'):
		super(Regressor,self).__init__()
		base_feat = 64
		n = 4
		self.regressor_set = []
		for _ in range(n_choices):
			self.regressor_set.append(
				nn.Sequential(
					nn.Linear(latent_dim,base_feat*n),
					nn.LeakyReLU(),
					Reshape([-1,n,base_feat]),
					nn.InstanceNorm1d(n,affine=True),
					Reshape([-1,n*base_feat]),
					
					nn.Linear(base_feat*n,base_feat*n),
					nn.LeakyReLU(),
					Reshape([-1,n,base_feat]),
					nn.InstanceNorm1d(n,affine=True),
					Reshape([-1,n*base_feat]),
						
					nn.Linear(base_feat*n,base_feat*n),
					nn.LeakyReLU(),
					Reshape([-1,n,base_feat]),
					nn.InstanceNorm1d(n,affine=True),
					Reshape([-1,n*base_feat]),
			
					nn.Linear(n*base_feat,n_confounds),
					#nn.ReLU(),
					Reshape([-1,n_confounds,1]),
					nn.Sigmoid()
				)
			)
	def parameters(self):
		return itertools.chain(*[r.parameters() \
			for r in self.regressor_set])
	def cuda(self,device):
		for r in self.regressor_set:
			r.cuda(device)
	def cpu(self):
		for r in self.regressor_set:
			r.cpu()
	def forward(self,x):
		return torch.cat([r(x) for r in self.regressor_set],2)
		
class Classifier(nn.Module):
	def __init__(self,latent_dim,n_inputs,base_feat,n_out,n_labels):
		super(Classifier,self).__init__()
		
		self.classifier = nn.Sequential(
			nn.Linear(latent_dim*n_inputs,n_inputs*base_feat*4),
			nn.LeakyReLU(),
			Reshape([-1,n_inputs,base_feat*4]),
			nn.InstanceNorm1d(n_inputs,affine=True),
			Reshape([-1,1,n_inputs*base_feat*4]),

			nn.Linear(in_features = n_inputs*base_feat*4,
				out_features = n_inputs*base_feat*2),
			nn.LeakyReLU(),
			Reshape([-1,n_inputs,base_feat*2]),
			nn.InstanceNorm1d(n_inputs,affine=True),
			Reshape([-1,1,n_inputs*base_feat*2]),

			nn.Linear(in_features = n_inputs*base_feat*2,
				out_features = n_out*n_labels),
			Reshape([-1,n_out,n_labels]),
			nn.Sigmoid(),	
		)
	def parameters(self):
		return self.classifier.parameters()
	def forward(self,x):
		return self.classifier(x)

class MultiInputModule(nn.Module):
	def __init__(self,
				Y_dim: tuple = (1,32), # Number of labels, Number of choices
				C_dim: tuple = (16,32), # Number of labels, Number of choices
				n_dyn_inputs: int = 14,
				n_stat_inputs: int = 2,
				use_attn: bool = False,
				encode_age: bool = False,
				variational: bool = False, # Makes it a variational encoder
				zero_input: bool = False, # Repeats input into classifier or makes it zeros
				remove_uncertain: bool = False,
				device = torch.device('cpu'),
				latent_dim: int = 128,
				weights: bool = None):
		super(MultiInputModule,self).__init__()
		
		# Model Parameters
		self.latent_dim = latent_dim
		
		# Number of non-image, patient-specific demographic inputs. Sex and 
		# ethnicity are two that can be applied.
		self.n_stat_inputs = n_stat_inputs
		
		# Max number of images that can be passed in
		self.n_dyn_inputs = n_dyn_inputs
		
		# Total number of inputs, which is used for the classifier model
		self.n_inputs = self.n_stat_inputs + self.n_dyn_inputs
		
		if (not isinstance(Y_dim,tuple)) or (len(Y_dim) != 2) or\
			(not isinstance(C_dim,tuple)) or (len(C_dim) != 2):
			raise Exception("""
				Y_dim and C_dim must be tuples.
				(<# labels>,<max # choices>)
				One label with two choices: (1,2)
			""")
		self.Y_dim = Y_dim
		self.C_dim = C_dim
		self.zero_input = zero_input
		self.remove_uncertain = remove_uncertain

		# Under development
		if self.remove_uncertain:
			self.record_training_sample = False
			self.num_training_samples = 300
			self.training_sample = torch.zeros(
				(
					self.latent_dim,
					self.num_training_samples
				),
				device=device)

		# Training options
		self.use_attn = use_attn
		self.encode_age = encode_age
		self.static_dropout = True # Randomly mask static inputs in training
		self.variational = variational
		
		# A record that prevents unrecognized keys from being applied during
		# the test phase
		self.static_record = [set() for _ in range(self.n_stat_inputs)]
		
		# Sets the multiplier for the number of features in each model component
		base_feat = 64
		# Modules
		
		# Makes the encoder output a variational latent space, so it's a
		# Gaussian distribution.
		if self.variational:
			self.encoder = Encoder(latent_dim=self.latent_dim)
			self.z_mean = nn.Sequential(
				nn.Linear(2*self.latent_dim,self.latent_dim)
			)
			self.z_log_sigma = nn.Sequential(
				nn.Linear(2*self.latent_dim,self.latent_dim)
			)
			self.epsilon = torch.distributions.Normal(0, 1)
			self.epsilon.loc = self.epsilon.loc.cuda(device)
			self.epsilon.scale = self.epsilon.scale.cuda(device)
		else:
			self.encoder = Encoder(latent_dim=self.latent_dim)
		# The output of the classifier and regressor, and encoder are kept
		# consistent, to 16 max outputs and 32 possible choices. This makes
		# cross-training easier, though it's less efficient.
		self.classifier = Classifier(latent_dim = self.latent_dim,
										n_inputs = self.n_inputs,
										base_feat = base_feat,
										n_out = self.Y_dim[0],#16,
										n_labels = self.Y_dim[1])#32)
		if self.C_dim is not None:
			n_confounds,n_choices = self.C_dim
			self.regressor = Regressor(self.latent_dim,
				n_confounds=self.C_dim[0],
				n_choices=self.C_dim[1],
				device=device)
		else: self.regressor = None
		
		if weights is not None:
			if self.C_dim != (16,32) or self.Y_dim != (1,32):
				warnings.warn(
					"Pretrained models may not function if defaults are altered"
					)
			if os.path.isfile(weights) and \
				os.path.splitext(weights)[1] == ".pt":
				self.load_state_dict(torch.load(weights))
			else:
				self.load_state_dict(torch.load(download_weights(weights)))

	def forward_ensemble(self,kwargs,n_ens=10):
		x = []
		for i in range(n_ens):
			x.append(self(**kwargs))
		return x

	def cuda(self,device):
		self.device=device
		if self.variational:
			self.epsilon.loc = self.epsilon.loc.cuda(device)
			self.epsilon.scale = self.epsilon.scale.cuda(device)
		self.regressor.cuda(device)
		return super().cuda(device)
	def cpu(self):
		self.device = torch.device('cpu')
		if self.variational:
			self.epsilon.loc = self.epsilon.loc.cpu()
			self.epsilon.scale = self.epsilon.scale.cpu()
		self.regressor.cpu()
		return super().cpu()
	
	def regressor_freeze(self):
		for param in self.classifier_parameters():
			param.requires_grad = True
		for param in self.regressor.parameters():
			param.requires_grad = False
			
	def classifier_freeze(self):
		for param in self.classifier_parameters():
			param.requires_grad = False
		for param in self.regressor.parameters():
			param.requires_grad = True

	def classifier_parameters(self):
		if self.variational:
			return itertools.chain(self.encoder.parameters(),
				self.classifier.parameters(),
				self.z_log_sigma.parameters(),
				self.z_mean.parameters())
		else:
			return itertools.chain(self.encoder.parameters(),
				self.classifier.parameters())

	def forward(self,
				x,
				static_input=None,
				dates=None,
				bdate=None,
				return_regress = False,
				return_encoded = False,
				encoded_input = False):
		dates1 = None
		bdate1 = None
		use_regression = hasattr(self,'regressor') and \
			(self.regressor is not None) and \
			(self.training or return_regress)
		if not encoded_input:
			if isinstance(x,BatchRecord):
				assert(x.dtype == "torch")
				if len(x.get_static_inputs()[0]) > 0:
					static_inputs = [_[0] for _ in x.get_static_inputs()]
					if self.n_stat_inputs != len(static_inputs):
						raise Exception(
					"Number of static inputs not equal to input: %d != %d"\
						 % (len(static_inputs),self.n_stat_inputs))
				dates1 = x.get_exam_dates()
				bdates = x.get_birth_dates()
				for i,b in enumerate(bdates):
					if b is not None:
						bdate1 = b
				x = x.get_image()
				assert(torch.is_tensor(x))
				
			if (self.encode_age and (dates is None or bdate is None)):
				raise Exception("Need dates as input to encode age")
			if x.size(0) > self.n_dyn_inputs:
				raise Exception(
					"Max dynamic inputs is %d. Received %d. Reduce batch size." % \
						(
							self.n_dyn_inputs,
							int(x.size(0))
						)
					)
			if static_input is not None and len(static_input) > self.n_stat_inputs:
				raise Exception(
					"Max dynamic inputs is %d. Received %d. Reduce batch size." % \
						(
							self.n_stat_inputs,
							len(static_input)
						)
					)
			# Encode everything - separate batches
			if self.variational:
				x = self.encoder(x)
				z_mean = self.z_mean(x)
				z_log_sigma = self.z_log_sigma(x)
				x = z_mean + (z_log_sigma.exp()*self.epsilon.sample(z_mean.shape))
				self.kl = (z_mean**2 + z_log_sigma.exp()**2 - z_log_sigma-0.5).mean()
			else:
				x = self.encoder(x) # [1-16]*96*96*96 -> [1-16]*512
			
			if hasattr(self,'remove_uncertain'):
				if self.remove_uncertain:
					if self.record_training_sample:
						self.training_sample[:,
							self.training_i:min(self.training_i + x.shape[0],
							self.num_training_samples)] = x
						self.training_i += x.shape[0]
						if self.training_i >= self.num_training_samples:
							self.record_training_sample = False
				
			if return_encoded:
				return x
		else:
			dates1 = dates
			bdate1 = bdate
		if use_regression:
			reg = self.regressor(x)
			# Encode dynamic inputs with dates using positional encoding
		if (self.encode_age) or \
			(self.training and self.static_dropout and \
				random.choice([True,False])):
			age_encodings = []
			if dates1 is not None:
				for i,date in enumerate(dates1):
					age_encoding = get_age_encoding(
									date,
									bdate1,
									d=self.latent_dim)
					age_encodings.append(age_encoding)
				age_encodings = np.array(age_encodings)
				age_encodings =  torch.tensor(
									age_encodings,
									device=x.device
									).float()
				x = torch.add(x,age_encodings)

		x = torch.unsqueeze(x,0)
		# Pad encodings with zeros, depending on input size
		e_size = list(x.size())
		e_size[1] = self.n_inputs - e_size[1] #[1-16]*512 -> 16*512
		if (not hasattr(self,'zero_input')) or self.zero_input:
			x = torch.cat((x,torch.zeros(e_size,device=x.device)),axis=1)
		else:
			x_ = torch.clone(x)
			while x_.size()[1] < e_size[1]:
				x_ = torch.cat((x_,torch.clone(x_)),axis=1)
			x_ = x_[:,:e_size[1],:]
			x = torch.cat((x,x_),axis=1)
			#x = x.repeat(1,(e_size[1]*2 // x.size()[1]),1)[:,:e_size[1],:]
		
		# Place static inputs near end
		if static_input is not None:
			if len(static_input) != self.n_stat_inputs:
				raise Exception(
						"Received %d static inputs, but it is set at %d" % \
						(len(static_input),self.n_stat_inputs)
					)
			if self.training:
				for i,e in enumerate(static_input):
					self.static_record[i].add(e)
			else:
				for i,e in enumerate(self.static_record):
					if static_input[i] not in e:
						raise Exception("Input %s not a previous demographic input (previous inputs were %s)" % (static_input[i],str(e)))
			x_ = encode_static_inputs(static_input,d=self.latent_dim)
			x_ = torch.tensor(x_,device = x.device)
			x_ = torch.unsqueeze(x_,0)
			for i in range(x_.shape[0]):
				if ((not self.static_dropout) or random.choice([True,False]) and self.training) or\
					(not self.static_dropout):
					x[:,(-(self.n_stat_inputs) + i):,:] = x_[i,:]
		
		# Randomly order dynamic encodings
		r = list(range(self.n_inputs))
		r_ = r[:self.n_stat_inputs]
		random.shuffle(r_)
		r[:self.n_stat_inputs] = r_
		x = x[:,r,...]
		
		# Apply attention mask
		if self.use_attn:
			m = torch.cat(
					(
						torch.zeros(
							list(x.size())[:],
							device=x.device,
							dtype=torch.bool),
						torch.ones(
							e_size[:],
							device=x.device,
							dtype=torch.bool)
					),
				axis=1)
			m = m[:,r,...]
			x,_ = self.multihead_attn(x,x,x,need_weights=False)#,attn_mask=m)

		# Switch batch channel with layer channel prior to running classifier
		x = torch.unsqueeze(x,-1)
		x = x.contiguous().view([-1,1,self.latent_dim*self.n_inputs]) # 16*512 -> 1*[16*512]
		x = self.classifier(x)
		if use_regression: return x,reg
		else: return x

class EnsembleModel(nn.Module):
	def __init__(self,model_list):
		super(EnsembleModel,self).__init__()
		self.models = [torch.load(_) for _ in model_list]
		for model in self.models: model.eval()
	def forward(self,input,hidden):
		new_output = []
		new_hidden = []
		for i in range(hidden.size()[3]):
			model = self.models[i]
			x = model.Encoder(input)
			x = torch.unsqueeze(x,1).float()
			output,h = model.RNN(x,hidden[...,i])
			new_output.append(torch.unsqueeze(output,3))
			new_hidden.append(torch.unsqueeze(h,3))
		new_output = torch.cat(new_output,dim=3)
		new_hidden = torch.cat(new_hidden,dim=3)
		return new_output,new_hidden



from options.train_options import *

import os
import sys
import time
wd = os.path.dirname(os.path.realpath(__file__))

from multi_med_image_ml.MedImageLoader import *
from multi_med_image_ml.models import *
from multi_med_image_ml.MultiInputTrainer import *
# Path function, used when switching platforms.

def key_to_filename(key,reverse=False):
	prefix = os.path.join(os.path.dirname(wd),'MGH_ML_pipeline')
	suffix = '_resized_96.npy'
	if not reverse:
		string = os.path.join(prefix,key[1:] + suffix)
	else:
		string = key[len(prefix):-len(suffix)]
	return string

opt = TrainOptions().parse()

model = MultiInputModule(weights=opt.pretrained_model)
model.cuda(opt.gpu_ids[0])

medim_loader = MedImageLoader(opt.all_vars,
	dim=opt.dim,
	cache=True,
	label=opt.label,
	confounds=opt.confounds,
	return_obj = True,
	dtype="torch",
	val_ranges=opt.val_ranges,
	static_inputs=opt.static_inputs,
	augment=opt.augment,
	key_to_filename=key_to_filename,
	gpu_ids = opt.gpu_ids
	)

trainer = MultiInputTrainer(model,
		batch_size=opt.batch_size,
		lr=opt.lr,
		name=opt.name,
		checkpoint_dir=opt.checkpoint_dir,
		loss_image_dir=os.path.join(opt.results_dir,
						opt.name,"loss_ims")
	)

for i in range(opt.epochs):
	print(f"Epoch {i}")
	j=0
	t = time.time()
	for p in medim_loader:
		j += 1
		trainer.loop(p,dataloader=medim_loader)
	t = time.time() - t
	print(f"Iters: {j}")
	print(f"Time: {t} seconds")



import pandas as pd
import os
import numpy as np
from .utils import *
from .Records import ImageRecord
import dateutil

class DataBaseWrapper():
	"""
	Wrapper for Pandas table to cache some common and repeated functions
	"""
	def __init__(self,
					all_vars = None,
					filename=None,
					labels=[],
					confounds=[],
					dim=None,
					cdim=None,
					key_to_filename = key_to_filename_default,
					val_ranges={}):
		self.key_to_filename = key_to_filename
		check_key_to_filename(self.key_to_filename)
		self.filename = filename
		self.dim = dim
		self.val_ranges = val_ranges
		self.labels = [] if labels is None else labels
		self.confounds = [] if confounds is None else confounds
		if all_vars is not None:
			self.all_vars = all_vars
			self.columns = set(self.all_vars.columns)
		elif os.path.isfile(filename):
			self.all_vars = pd.read_pickle(self.filename)
			self.columns = set(self.all_vars.columns)
		elif self.filename is not None:
			assert(self.dim is not None)
			self.all_vars = pd.DataFrame()
			self.columns = set()
		else:
			raise Exception("DatabaseWrapper cannot have no inputs")
		self.jdict = []
		
		if isinstance(self.all_vars,str) and os.path.isfile(self.all_vars)\
			 and os.path.splitext(self.all_vars)[1] == ".pkl":
			self.all_vars = pd.read_pickle(self.all_vars)
	def has_im(self,im: ImageRecord):
		fkey = self.key_to_filename(im.npy_file,reverse=True)
		return fkey in self.all_vars.index
	def in_val_ranges(self,fkey: str) -> bool:
		valid = True
		for column_name in self.val_ranges:
			val = self.all_vars.loc[fkey,column_name]
			if is_nan(val,inc_null_str=True):
				return False
			elif isinstance(val,str):
				if val not in self.val_ranges[column_name]:
					return False
			elif isinstance(val,float) or isinstance(val,int):
				min_,max_ = self.val_ranges[column_name]
				if val < min_ or val > max_:
					return False
		return True
	def build_metadata(self):
		for l in self.labels:
			if l not in self.all_vars.columns:
				print("%s not in all_vars.columns" % l)
				print(self.all_vars.columns)
			assert(l in self.all_vars.columns)
		for c in self.confounds:
			assert(c in self.all_vars.columns)
		assert(len(set(self.labels).intersection(set(self.confounds)))==0)
		self.uniques = {}
		self.n_buckets=10
		for c in self.confounds + self.labels:
			self.uniques[c] = {}
			lis = list(self.all_vars.loc[:,c])
			if np.any([isinstance(_,str) for _ in lis]):
				self.uniques[c]["discrete"] = True
				u = {}
				for l in lis:
					if not is_nan(l):
						if l not in u:
							u[l] = 0
						u[l] += 1
				# This ensures that confounds are ordered by their frequency
				# of appearance in the real data. Thus, if there were 7500
				# females and 2500 males in a dataset, females would be
				# encoded as [1,0] and males as [0,1].
				if c in self.confounds:
					u = sorted(
								[(_,u[_]) for _ in u],
								key=lambda k: k[1],
								reverse=True)
					u = [_[0] for _ in u]
				else:
					try:
						u = sorted(list(u))
					except:
						print(u)
						exit()
				self.uniques[c]["unique"] = u
				self.n_buckets = max(self.n_buckets,len(u))
			else:
				self.uniques[c]["discrete"] = False
				max_ = -np.inf
				min_ = np.inf
				nonnan_list = []
				for l in lis:
					if not is_nan(l):
						max_ = max(max_,l)
						min_ = min(min_,l)
						nonnan_list.append(l)
				self.uniques[c]["max"] = max_
				self.uniques[c]["min"] = min_
				self.uniques[c]["nonnan_list"] = sorted(nonnan_list)

				self.n_buckets_cont = min([self.n_buckets,
									10,
									int(len(self.uniques[c]["nonnan_list"])/10)
								])
				skips = int(len(self.uniques[c]["nonnan_list"])/\
					float(self.n_buckets_cont)) + 1
				self.uniques[c]["nonnan_list"] = \
					self.uniques[c]["nonnan_list"][::skips]
				# Get mean between density and range dists
				max_ = self.uniques[c]["max"]
				min_ = self.uniques[c]["min"]
				rd = np.arange(self.n_buckets_cont)
				rd = rd / float(self.n_buckets_cont-1)
				rd = list((rd * (max_ - min_)) + min_)
				self.uniques[c]["nonnan_list"] = \
					[(rd[i] + \
					self.uniques[c]["nonnan_list"][i])/2 \
					for i in range(self.n_buckets_cont)]
				self.uniques[c]["nonnan_list"][-1] = max_
				self.uniques[c]["nonnan_list"][0] = min_
				assert(len(self.uniques[c]["nonnan_list"]) == \
					self.n_buckets_cont)
	def get_confound_encode(self,npy_file: str):
		assert(os.path.splitext(npy_file)[1] == ".npy")
		confound_strs = [self.loc_val(npy_file,c) for c in self.confounds]
		cnum_list = []
		for j,c in enumerate(self.confounds):
			if self.uniques[c]["discrete"]:
				c_uniques = self.uniques[c]["unique"]
				if is_nan(confound_strs[j]):
					cnum_list.append(-1)
				else:
					cnum_list.append(c_uniques.index(confound_strs[j]))
			else:
				max_ = self.uniques[c]["max"]
				min_ = self.uniques[c]["min"]
				if is_nan(confound_strs[j]):
					cnum_list.append(-1)
				else:
					unnl = self.uniques[c]["nonnan_list"]
					for kk in range(len(unnl)-1):
						if unnl[kk] <= confound_strs[j] and \
							unnl[kk+1] >= confound_strs[j]:
							cnum_list.append(kk)
							break
		assert(len(cnum_list) == len(self.confounds))
		return cnum_list
	def get_label_encode(self,npy_file: str):
		assert(os.path.splitext(npy_file)[1] == ".npy")
		label_strs = [self.loc_val(npy_file,c) for c in self.labels]
		cnum_list = []
		for j,c in enumerate(self.labels):
			if self.uniques[c]["discrete"]:
				c_uniques = self.uniques[c]["unique"]
				if is_nan(label_strs[j]):
					cnum_list.append(-1)
				else:
					cnum_list.append(c_uniques.index(label_strs[j]))
			else:
				max_ = self.uniques[c]["max"]
				min_ = self.uniques[c]["min"]
				if is_nan(confound_str[j]):
					cnum_list.append(-1)
				else:
					unnl = self.uniques[c]["nonnan_list"]
					for kk in range(len(unnl)-1):
						if unnl[kk] <= label_strs[j] and \
							unnl[kk+1] >= label_strs[j]:
							cnum_list.append(kk)
							break
		assert(len(cnum_list) == len(self.labels))
		return cnum_list
	def _get_val(self,npy_file: str,potential_columns):
		assert(np.all([isinstance(_,str) for _ in potential_columns]))
		val = None
		#fkey = get_dim_str(filename,dim=self.dim)
		for c in potential_columns:
			if c in self.columns:
				val = self.loc_val(npy_file,c)
				if not(is_nan(val) or isinstance(val,str)):
					print(val)
				assert(is_nan(val) or isinstance(val,str))
				if not is_nan(val):
					break
		
		return val
	def get_ID(self,npy_file):
		id = self._get_val(npy_file,["PatientID","Patient ID"])
		return id
	def parse_date(self,d,date_format="%Y-%m-%d %H:%M:%S"):
		if is_nan(d,inc_null_str=True):
			return datetime.datetime(year=1970,month=1,day=1)
		elif is_float(d):
			return dateutil.parser.parse(str(d))
		else:
			try:
				return datetime.datetime.strptime(
					d.replace("_"," ").split(".")[0],
					date_format)
			except ValueError:
				return dateutil.parser.parse(d.replace("_"," "))
	def get_exam_date(self,npy_file: str) -> datetime.date:
		d = self._get_val(npy_file,["ExamEndDTS","Acquisition Date"])
		return self.parse_date(d)
	def get_birth_date(self,npy_file: str) -> datetime.date:
		d = self._get_val(npy_file,["BirthDTS"])
		return self.parse_date(d)
	
	def loc_val(self,npy_file,c):
		fkey = self.key_to_filename(npy_file,reverse=True)
		try:
			#if fkey not in self.all_vars.index:
			#	print("Error - %s not present in index. Adding..." % fkey)
			#assert(fkey in self.all_vars.index)
			
			return self.all_vars.loc[fkey,c]
		except KeyError:
			nifti_file = get_dim_str(fkey,dim=self.dim,outtype=".nii.gz")
			
			if not os.path.isfile(nifti_file):
				nifti_file = get_dim_str(fkey,dim=self.dim,outtype=".nii")
			if not os.path.isfile(nifti_file):
				raise Exception("Key error: %s" % nifti_file)
			json_file = date_sorter(os.path.dirname(nifti_file),".json")[0]
			if not os.path.isfile(json_file):
				raise Exception("Key error - no JSON: %s" % fkey)
			self.add_json(nifti_file = nifti_file,json_file=json_file)
			self.out_dataframe(fkey_ass=fkey)
			return self.all_vars.loc[fkey,c]
	def add_json(self,nifti_file,json_file=None):
		if not not_temp(nifti_file): return
		if json_file is None:
			json_file = "%s.json" % os.path.splitext(
				os.path.splitext(nifti_file)[0])[0]
			if not (os.path.isfile(json_file)):
				json_file = os.path.join(
					os.path.dirname(nifti_file),
					"metadata.json")
				if not os.path.isfile(json_file):
					return
		npy_file = get_dim_str(nifti_file,self.dim)
		fkey = self.key_to_filename(npy_file,reverse=True)
		if fkey not in self.jdict and fkey not in self.all_vars.index:
			with open(json_file,'r') as fileobj:
				json_dict = json.load(fileobj)
			json_dict["fkey"] = fkey
			json_dict["filename"] = fkey
			for item in json_dict:
				if isinstance(json_dict[item],list):
					json_dict[item] = "_".join(
							[str(_) for _ in sorted(json_dict[item])]
						)
			self.columns = self.columns.union(set(json_dict))
			self.jdict.append(json_dict)
			assert(len(self.jdict) > 0)
	def get_file_list(self):
		flist = list(filter(lambda k: self.in_val_ranges(k),
				self.all_vars.index))
		return [self.key_to_filename(str(_)) for _ in flist]
	def out_dataframe(self,fkey_ass = None):
		
		if len(self.jdict) > 0:
			out = pd.DataFrame(self.jdict,columns = list(self.columns))
			out.set_index("fkey",inplace=True)
			if len(self.all_vars) > 0:
				self.all_vars = pd.concat([self.all_vars,out],
						ignore_index=False,
						join='inner')
				if fkey_ass is not None: assert(fkey_ass in self.all_vars.index)
			else:
				self.all_vars = out
				if fkey_ass is not None: assert(fkey_ass in self.all_vars.index)
			assert(len(self.all_vars) > 0)
			if fkey_ass is not None: assert(fkey_ass in self.all_vars.index)
			self.all_vars.drop_duplicates(inplace=True)
			if fkey_ass is not None: assert(fkey_ass in self.all_vars.index)
			self.all_vars.to_pickle(self.filename)
			if fkey_ass is not None: assert(fkey_ass in self.all_vars.index)
			self.jdict = []
	def stack_list_by_label(self,filename_list,label):
		if label in self.labels:
			sel_list = self.labels
		elif label in self.confounds:
			sel_list = self.confounds
		else:
			raise Exception("%s not in labels or confounds")
		lnum = sel_list.index(label)
		if label in self.labels:
			lencodes = [self.get_label_encode(f)[lnum] for f in filename_list]
		elif label in self.confounds:
			lencodes = [self.get_confound_encode(f)[lnum] for f in filename_list]
		filename_list_stack = [[] for _ in range(len(np.unique(lencodes)))]
		for filename,l in zip(filename_list,lencodes):
			filename_list_stack[l].append(filename)
		return filename_list_stack
